﻿using Microsoft.Web.Services3;
using Microsoft.Web.Services3.Security;
using Microsoft.Web.Services3.Security.Tokens;
using System;
using System.IO;
using System.Security.Cryptography;
using System.Security.Cryptography.X509Certificates;
using System.Security.Cryptography.Xml;
using System.Text;
using System.Xml;
using System.Xml.XPath;

namespace ConsoleApp1
{
    internal class ClientOutputFilter : SoapFilter
    {
        private readonly Security security;

        public ClientOutputFilter(SigningPolicyAssertion parentAssertion, Security security)
        {
            this.security = security;
        }
        public override SoapFilterResult ProcessMessage(SoapEnvelope envelope)
        {
            // Remove unnecessary XML elements
            var actionNode = envelope.Header.GetElementsByTagName("Action", "http://schemas.xmlsoap.org/ws/2004/08/addressing");
            var messageIdNode = envelope.Header.GetElementsByTagName("MessageID", "http://schemas.xmlsoap.org/ws/2004/08/addressing");
            var replyToNode = envelope.Header.GetElementsByTagName("ReplyTo", "http://schemas.xmlsoap.org/ws/2004/08/addressing");
            var toNode = envelope.Header.GetElementsByTagName("To", "http://schemas.xmlsoap.org/ws/2004/08/addressing");
            envelope.Header.RemoveChild(actionNode.Item(0));
            envelope.Header.RemoveChild(messageIdNode.Item(0));
            envelope.Header.RemoveChild(replyToNode.Item(0));
            envelope.Header.RemoveChild(toNode.Item(0));

            // Attach an ID attribute to the Body element. This is required for the References.
            XmlAttribute idAttrib = envelope.CreateAttribute("wsu", "Id", "http://docs.oasis-open.org/wss/2004/01/oasis-200401-wss-wssecurity-utility-1.0.xsd");
            idAttrib.Value = "Body";
            envelope.Body.Attributes.Append(idAttrib);

            // Read the signing certificate
            var keyPath = "alairo.p12";
            var keyPassword = "1111";
            X509Certificate2 certificate = new X509Certificate2(keyPath, keyPassword);

            // WSSE Security element
            var token = new X509SecurityToken(certificate);
            security.Tokens.Add(token);
            security.Timestamp.TtlInSeconds = 43200;
            security.SerializeXml(envelope);

            // Digital signature
            var signedXml = new SignedXmlWithId(envelope);
            signedXml.SigningKey = certificate.GetRSAPrivateKey();
            signedXml.SignedInfo.SignatureMethod = SignedXml.XmlDsigRSASHA384Url;
            signedXml.SignedInfo.CanonicalizationMethod = SignedXml.XmlDsigExcC14NTransformUrl;

            AddReference(signedXml, security.Timestamp.Id);
            AddReference(signedXml, "Body");

            KeyInfo keyInfo = new KeyInfo();
            SecurityTokenReference tokenRef = new SecurityTokenReference(token);
            keyInfo.AddClause(tokenRef);
            signedXml.KeyInfo = keyInfo;

            signedXml.ComputeSignature();
            XmlElement signedElement = signedXml.GetXml();

            // Insert the signature into the WSSE Security element of the SOAP Header
            XPathNavigator signatureNavigator = signedElement.CreateNavigator();
            XPathNavigator headerNavigator = envelope.Header.CreateNavigator();
            XmlNamespaceManager namespaceManager = new XmlNamespaceManager(envelope.NameTable);
            namespaceManager.AddNamespace("wsse", "http://docs.oasis-open.org/wss/2004/01/oasis-200401-wss-wssecurity-secext-1.0.xsd");
            XPathNavigator secNode = headerNavigator.SelectSingleNode("wsse:Security", namespaceManager);
            secNode.AppendChild(signatureNavigator);

            return SoapFilterResult.Continue;
        }

        private void AddReference(SignedXml signedXml, string uri)
        {
            Reference reference = new Reference();
            reference.Uri = "#" + uri;
            reference.DigestMethod = SignedXml.XmlDsigSHA384Url;

            XmlDsigExcC14NTransform env = new XmlDsigExcC14NTransform();
            reference.AddTransform(env);

            signedXml.AddReference(reference);
        }
    }
}