﻿using Org.BouncyCastle.Security;
using System;
using System.IO;
using Org.BouncyCastle.Asn1.Pkcs;
using Org.BouncyCastle.Asn1.X509;
using Org.BouncyCastle.Crypto;
using Org.BouncyCastle.Crypto.Generators;
using Org.BouncyCastle.Crypto.Operators;
using Org.BouncyCastle.OpenSsl;
using Org.BouncyCastle.Pkcs;
using Org.BouncyCastle.X509;
using System.Security.Cryptography.X509Certificates;
using System.Security.Cryptography;
using System.Text;
using System.Linq;

namespace ConsoleApp1
{
    class Program
    {
        static void Main(string[] args)
        {
            Console.WriteLine("NTAK Tester");

            /* The below two lines can be used to generate CSR files. Only need to be run once. */
            //generateKeyPairAndCsr("auth");
            //generateKeyPairAndCsr("alairo");

            /* Store the certificates in the key store. Only need to be run once. */
            //store("auth");
            //store("alairo");

            /* Example call to NTAK */
            napiZarasUtemezes();

            //saltTest();
        }

        static void generateKeyPairAndCsr(string type)
        {
            // Generation of 4096-bit RSA key pair
            var random = new SecureRandom();
            var parameters = new KeyGenerationParameters(random, 4096);
            var generator = new RsaKeyPairGenerator();
            generator.Init(parameters);
            var keypair = generator.GenerateKeyPair();

            FileStream file = new FileStream(type + ".pem", FileMode.Create, FileAccess.Write, FileShare.Read);
            StreamWriter streamWriter = new StreamWriter(file);
            PemWriter pemWriter = new PemWriter(streamWriter);
            pemWriter.WriteObject(keypair);
            streamWriter.Close();

            // Fill in the appropriate values for your accommodation
            string szallasRegisztraciosSzam = "nbI3z29w";
            string szallasNev = "PMS_Szmetanko_Hotel";
            string szallasIranyitoszam = "1023";
            string szallasTelepules = "Budapest";
            string szallashelySzolgaltatoAdoszam = "11111111-9-01";

            // Generate certificate request
            //
            // The value of the subjectName field has to be populated with the accommodation’s details
            // Common Name (CN) – The name of the Subject: The accommodation’s registration number
            // Organization (O) – The name of the Organization: Accommodation name
            // Organization Identifier (OrgId) – OID: 2.5.4.97 – Organisation ID: The accommodation services provider’s tax number.
            // Country (C) – Country identifier: Hungary, i.e. the ‘HU’ value is to be entered here in every case
            // Locality Name (L) – Settlement name: Accommodation settlement
            // Postal Code – OID: 2.5.4.17 – Postcode: Accommodation postcode

            var subjectName = $"CN={szallasRegisztraciosSzam}, O={szallasNev}, 2.5.4.97={szallashelySzolgaltatoAdoszam}, 2.5.4.17={szallasIranyitoszam}, L={szallasTelepules}, C=HU";
            var subject = new X509Name(subjectName);
            var factory = new Asn1SignatureFactory(PkcsObjectIdentifiers.Sha512WithRsaEncryption.Id, keypair.Private, random);
            var request = new Pkcs10CertificationRequest(factory, subject, keypair.Public, null, keypair.Private);

            file = new FileStream(type + ".csr", FileMode.Create, FileAccess.Write, FileShare.Read);
            streamWriter = new StreamWriter(file);
            pemWriter = new PemWriter(streamWriter);
            pemWriter.WriteObject(request);
            streamWriter.Close();
        }

        static void store(string type)
        {
            // Read certificate
            X509CertificateParser parser = new X509CertificateParser();
            FileStream file = new FileStream(type + ".cer", FileMode.Open, FileAccess.Read, FileShare.Read);
            Org.BouncyCastle.X509.X509Certificate certificate = parser.ReadCertificate(file);
            file.Close();

            // Read key-pair
            file = new FileStream(type + ".pem", FileMode.Open, FileAccess.Read, FileShare.Read);
            StreamReader streamReader = new StreamReader(file);
            PemReader pemReader = new PemReader(streamReader);
            AsymmetricCipherKeyPair keypair = (AsymmetricCipherKeyPair)pemReader.ReadObject();
            streamReader.Close();

            // Create key container
            var random = new SecureRandom();
            string keyPassword = "1111";
            Pkcs12Store store = new Pkcs12StoreBuilder().Build();

            // Upload key container
            X509CertificateEntry certificateEntry = new X509CertificateEntry(certificate);
            store.SetCertificateEntry("nbI3z29w", certificateEntry);
            AsymmetricKeyEntry keyEntry = new AsymmetricKeyEntry(keypair.Private);
            store.SetKeyEntry("nbI3z29w", keyEntry, new X509CertificateEntry[] { certificateEntry });

            // Save key container
            file = new FileStream(type + ".p12", FileMode.Create, FileAccess.Write, FileShare.Read);
            store.Save(file, keyPassword.ToCharArray(), random);
            file.Close();
        }

        static void napiZarasUtemezes()
        {
            // Read authentication key
            String keyPath = "auth.p12";
            String keyPassword = "1111";
            X509Certificate2 certificate = new X509Certificate2(keyPath, keyPassword);

            // Client class structure
            var client = new NTAK.napiZarasUtemezesPortv8Soap11();
            client.Url = "https://pms.sgmdev.hu/ntak";
            client.ClientCertificates.Add(certificate);

            var szallashely = new NTAK.SzallasType();
            szallashely.szallasAdatszolgaltatasiAzonosito = "nbI3z29w";
            szallashely.szallasNev = "PMS_Szmetanko_Hotel";
            szallashely.szallashelySzolgaltatoNev = "PMS_teszt Kft.";
            szallashely.szallashelySzolgaltatoAdoszam = "11111111-9-01";

            // Sample call
            NTAK.napiZarasUtemezesRequest request = new NTAK.napiZarasUtemezesRequest();
            request.szallashely = szallashely;
            request.uzenetId = System.Guid.NewGuid().ToString();
            request.szoftverVerzio = "v1.0.1";
            var response = client.napiZarasUtemezes(request);

            foreach (var kertAdat in response.kertAdatok)
            {
                foreach (var napiZaras in kertAdat.napiZarasok)
                {
                    Console.WriteLine(napiZaras.targynap);
                }
            }
            Console.ReadLine();
        }

        private static void saltTest()
        {
            // Personal data that needs to be hidden
            var szemelyesAdat = "Cement Ede Budapest 1969.10.29.";

            // Salt value applied
            string salt = "$2a$10$e0gw2jJ.bO5uLAAE5VNbH.";
            // Generate SHA-256 hash from personal data
            string vendegId = generateVendegId(szemelyesAdat, salt);
            Console.WriteLine(vendegId);
            Console.ReadLine();
        }

        private static string generateVendegId(string szemelyesAdat, string salt)
        {
            string hash = string.Concat(
                            new SHA256Managed().
                            ComputeHash(Encoding.UTF8.GetBytes(szemelyesAdat)).
                            Select(b => b.ToString("X2")).
                            ToArray());
            // BCrypt encryption based on the hash and salting
            string encrypted = BCrypt.Net.BCrypt.HashPassword(hash.ToLower(), salt);

            // The first 29 characters comprise the algorithm used, its other parameters, 
            // and the value of the salt, therefore have to be discarded before posting
            var vendegId = new string(encrypted.Skip(29).ToArray());
            return vendegId;
        }

    }
}
